import java.util.*;


/**
 * Simple command line option parser.
 *  This will handle simple switches such as '-b' or longer switches such
 *  as '--break=on'.
 */
public class CommandLineOptions
{
    /**
     * <CODE>Map</CODE> storing key value pairs for the input.
     */
    private Map      optionsMap;

    /**
     * Inidicates if <CODE>optionsMap</CODE> is upto date or needs
     *  regenerating.
     */
    private boolean  optionsMapDirty;

    /**
     * The actual options to parse.
     */
    private String[] options;

    /**
     * <CODE>Map</CODE> of command line switches, and the keys to which
     *  the values should be associated.
     */
    private Map      switchMap;

    /**
     * <CODE>Map</CODE> relating each possible key to an <CODE>Option</CODE>
     *  object containing the description of the switch and a <CODE>Set</CODE>
     *  of the switches that can be used to set key values.
     */
    private Map      switchInfo;

    /**
     * Create a CommandLineOptions object to represent the options
     *  parsed from some argument list.
     *@param args  The argument list to parse.
     */
    public CommandLineOptions(String args[])
    {
        switchMap       = new TreeMap();
        switchInfo      = new TreeMap();
        options         = args;
        optionsMapDirty = true;
    }


    /**
     * Add a single of option for some key.  This can be called multiple times
     *  as needed if more than one switch is required to control the value of
     *  a single key.
     *@param swtch  The first switch that sets a value to associate with <CODE>key</CODE>.
     *@param key  An object with which switch values will be associated, which should also
     *              be used to retrieve option values with the <CODE>get()</CODE>
     *              method.
     *@param description  Description of what is controlled by the switch.  This
     *                      is used for the switch descriptions provided by
     *                      <CODE>usage()</CODE>.
     */
    public void addOption(String swtch, Object key, String description)
    {
        Option o;

        /* Find or create an Option object for the key */
        o = (Option)switchInfo.get(key);
        if(o == null)
        {
            o = new Option(description);

            switchInfo.put(key, o);
        }

        /* Add the switch to the option info */
        o.addSwitch(swtch);

        switchMap.put(swtch, key);
        optionsMapDirty = true;
    }


    /**
     * Add a pair of options for the same key.  This can be used to conveniently
     *  set a long and short form of a command line switch where both switches
     *  control the same key value.
     *@param swtch0  The first switch that sets a value to associate with <CODE>key</CODE>.
     *@param swtch1  A second switch that sets a value to associate with <CODE>key</CODE>.
     *@param key  An object with which switch values will be associated, which should also
     *              be used to retrieve option values with the <CODE>get()</CODE>
     *              method.
     *@param description  Description of what is controlled by the switch.  This
     *                      is used for the switch descriptions provided by
     *                      <CODE>usage()</CODE>.
     */
    public void addOption(String swtch0,
                          String swtch1,
                          Object key,
                          String description)
    {
        Option o;

        /* Find or create an Option object for the key */
        o = (Option)switchInfo.get(key);
        if(o == null)
        {
            o = new Option(description);

            switchInfo.put(key, o);
        }

        /* Add the switches to the option info */
        o.addSwitch(swtch0);
        o.addSwitch(swtch1);

        /* Add the parameters for parsing the input */
        switchMap.put(swtch0, key);
        switchMap.put(swtch1, key);
        optionsMapDirty = true;
    }


    /**
     * Get the value for some command line option as a <CODE>String</CODE>.
     *@param key  A key that may have had some value associated with it.
     *@return Either a <CODE>String</CODE> being the supplied commandline value
     *         or <CODE>null</CODE> if the option was not specified.
     */
    public String get(String key)
    {
        if(optionsMapDirty)
        {
            parseOptions();
        }

        return (String)optionsMap.get(key);
    }


    /**
     * Display usage information for each command line switch.
     */
    public String usage()
    {
        StringBuffer sb    = new StringBuffer();
        boolean      first = true;
        Iterator     i;

        i = switchInfo.keySet().iterator();
        while(i.hasNext())
        {
            if(!first)
            {
                sb.append("\n");
            }
            else
            {
                first = false;
            }
            sb.append(switchInfo.get(i.next()));

        }

        return sb.toString();
    }


    /**
     * Return a <CODE>String</CODE> describing the options contained within
     *  this map.
     */
    public String toString()
    {
        StringBuffer sb = new StringBuffer("[");
        Iterator     i;

        if(optionsMapDirty)
        {
            parseOptions();
        }

        i = optionsMap.keySet().iterator();
        while(i.hasNext())
        {
            Object key = i.next();

            sb.append(key);
            sb.append("=");
            sb.append(optionsMap.get(key));
            sb.append(" ");
        }

        sb.append("]\n");

        return sb.toString();
    }


    /**
     * Parse the options input to populate the <CODE>optionsMap</CODE>.
     */
    private void parseOptions()
    {
        int t;

        optionsMap = new TreeMap();

        t = 0;
        while(t < options.length)
        {
            String swtch;
            int    eq;

            /* Check if the switch contains an equals sign, and get just
             *  the portion upto and including the '='
             */
            eq = options[t].indexOf('=');
            if(eq != -1)
            {
                swtch = options[t].substring(0, eq + 1);
            }
            else
            {
                swtch = options[t];
            }


            if(!switchMap.containsKey(swtch))
            {
                throw new RuntimeException("Unrecognised switch '" + swtch + "'.");
            }
            else
            {
                /* Get the key for the value */
                Object key = switchMap.get(swtch);
                String value;

                /* Now get the value */
                if(eq == -1)
                {
                    if(t + 1 >= options.length)
                    {
                        throw new RuntimeException("Switch '" + swtch + "' must have a value.");
                    }
                    value = options[t + 1];
                    t++;
                }
                else
                {
                    value = options[t].substring(eq + 1);
                }

                /* Add the value against the key */
                if(optionsMap.put(key, value) != null)
                {
                    throw new RuntimeException("Multiple values specified for '" + swtch + "'.");
                }
            }

            t++;
        }

        /* Just computed the map */
        optionsMapDirty = false;
    }


    /**
     * Class to record information about a single option.
     */
    private static class Option
    {
        /**
         * Human readable description of the option represented.
         */
        private String description;

        /**
         * The switches that can be used to control this opton.
         */
        private Set    switches = new TreeSet();


        /**
         * Create an option with some description.
         */
        Option(String description)
        {
            this.description = description;
        }


        /**
         * Add a switch to this option.
         */
        void addSwitch(String swtch)
        {
            switches.add(swtch);
        }


        /**
         * Return a <CODE>String</CODE> describing this option in a
         *  human readable form.  The returned string will be in the
         *  form of:
         *<PRE>
         *-a, -b, -c
         *  description
         *</PRE>
         */
        public String toString()
        {
            StringBuffer sb    = new StringBuffer("  ");
            boolean      first = true;
            Iterator     i;

            /* List the switches */
            i = switches.iterator();
            while(i.hasNext())
            {
                String option = (String)i.next();

                if(first)
                {
                    first = false;
                }
                else
                {
                    sb.append(", ");
                }

                sb.append(option);
                if(option.endsWith("="))
                {
                    sb.append("<value>");
                }
            }

            sb.append("\n  ");
            sb.append(description);
            sb.append("\n");

            return sb.toString();
        }

    }
}
